@@ -314,6 +314,7 @@
         #-sb-devel
         (warn 'cross-type-giving-up :call `(ctypep ,obj ,ctype)))))
 
+;; TODO: would it be feasible to unify this definition with that in src/code/typep ?
 (defun ctype-of (x)
   (typecase x
     (function
@@ -332,7 +333,7 @@
      (ctype-of-number x))
     (array
      ;; It is critical not to inquire of the host for the array's element type.
-     (let ((etype (specifier-type (sb-xc:array-element-type x))))
+     (let ((etype (specifier-type (array-element-type x))))
        (make-array-type (array-dimensions x)
                         ;; complexp relies on the host implementation,
                         ;; but in practice any array for which we need to
@@ -368,34 +369,11 @@
   (declare (ignore x))
   nil)
 
-;;; Every architecture needs this portable replacement for +-modfx.
-;;; Some, but not all, define +-modfx in cross-modular.
-;;; We need this available sooner than that because:
-;;;  - src/code/alien-type uses sb-xc:defstruct
-;;;  - sb-xc:defstruct needs to parse a MEMBER type because PARSE-1-DSD asks
-;;;    if (CSUBTYPEP CTYPE (SPECIFIER-type '(or fixnum boolean ...))).
-;;;  - BOOLEAN is defined as a MEMBER type.
-;;;  - MEMBER types go in a hashset.
-;;;  - The hashset's hash function by way of XSETs involves +-modfx.
-;;; And so there you go. I tried moving cross-modular earlier in the build; it failed.
-(defun plus-mod-fixnum (a b)
-  (declare (type sb-xc:fixnum a b))
-  (let* ((mask (ldb (byte sb-vm:n-fixnum-bits 0) -1))
-         (result (logand (+ (logand a mask) (logand b mask)) mask)))
-    (if (logbitp sb-vm:n-positive-fixnum-bits result) ; then it's negative
-        (dpb result (byte sb-vm:n-fixnum-bits 0) -1)
-        result)))
-
-;; Assume 2 fixnum tag bits:
-;;      significant bits      tag
-;;   #b01111...11111111111111  00
-;; + #b0                    1  00
-;;   ------------------------
-;; = #b10000...00000000000000  00
-(assert (= (plus-mod-fixnum sb-xc:most-positive-fixnum 1)
-           sb-xc:most-negative-fixnum))
-;; etc
-(assert (= (plus-mod-fixnum sb-xc:most-negative-fixnum sb-xc:most-negative-fixnum)
-           0))
-(assert (= (plus-mod-fixnum -1 most-negative-fixnum)
-           sb-xc:most-positive-fixnum))
+(defun non-null-symbol-p (x) (and x (symbolp x)))
+;; these two functions don't need to be fully general
+(defun pointerp (x)
+  (aver (or (symbolp x) (fixnump x)))
+  (symbolp x))
+;; Use of non-ASCII during build occurs no sooner than make-target-2,
+;; therefore _every_ character satisfies BASE-CHAR-P prior to that.
+#+sb-unicode (defun base-char-p (x) (characterp x))
