@@ -58,18 +58,18 @@
 ;;; These aren't so much "wrappers" as they are the actual installed DEFUNs.
 ;;; I supposed they "wrap" a vop or source-transform.
 (macrolet ((def-type-predicate-wrapper (pred)
-             (let* ((name (symbol-name pred))
-                    (stem (string-left-trim "%" (string-right-trim "P-" name)))
-                    (article (if (position (schar name 0) "AEIOU") "an" "a")))
-               `(defun ,pred (object)
-                  ;; Document the standardized predicates and not the internal ones.
-                  ,@(when (eql (sb-xc:symbol-package pred) *cl-package*)
+             `(defun ,pred (object)
+                ;; Document the standardized predicates and not the internal ones.
+                ,@(when (eq (sb-xc:symbol-package pred) *cl-package*)
+                    (let* ((name (symbol-name pred))
+                           (stem (string-left-trim "%" (string-right-trim "P-" name)))
+                           (article (if (position (schar name 0) "AEIOU") "an" "a")))
                       (list (format nil
                                     "Return true if OBJECT is ~A ~A, and NIL otherwise."
                                     article
-                                    stem)))
+                                    stem))))
                   ;; (falling through to low-level implementation)
-                  (,pred object)))))
+                (,pred object))))
   (def-type-predicate-wrapper array-header-p)
   (def-type-predicate-wrapper simple-array-header-p)
   (def-type-predicate-wrapper arrayp)
@@ -111,6 +111,8 @@
   #-(or x86 x86-64 arm64 riscv) (def-type-predicate-wrapper lra-p)
   (def-type-predicate-wrapper null)
   (def-type-predicate-wrapper numberp)
+  (sb-c::when-vop-existsp (:translate pointerp)
+    (def-type-predicate-wrapper pointerp))
   (def-type-predicate-wrapper rationalp)
   (def-type-predicate-wrapper ratiop)
   (def-type-predicate-wrapper realp)
@@ -156,11 +158,6 @@
   (def-type-predicate-wrapper stringp)
   (def-type-predicate-wrapper vectorp))
 
-(sb-c::when-vop-existsp (:translate fixnum-mod-p)
-  (defun fixnum-mod-p (x limit)
-    (and (fixnump x)
-         (<= 0 x limit))))
-
 (sb-c::when-vop-existsp (:translate car-eq-if-listp)
   (defun car-eq-if-listp (value object)
     (car-eq-if-listp value object)))
@@ -241,7 +238,6 @@
 
 (declaim (inline %eql))
 (defun %eql (obj1 obj2)
-  "Return T if OBJ1 and OBJ2 represent the same object, otherwise NIL."
   #+x86-64 (eql obj1 obj2) ; vop fully implements all cases of EQL
   #-x86-64 ; else this is the only full implementation of EQL
   (or (eq obj1 obj2)
@@ -285,6 +281,7 @@
                      (eql (imagpart x) (imagpart y))))))))))
 
 (defun eql (x y)
+  "Return T if OBJ1 and OBJ2 represent the same object, otherwise NIL."
   ;; On x86-64, EQL is just an interpreter stub for a vop.
   ;; For others it's a call to the implementation of generic EQL.
   (#+x86-64 eql #-x86-64 %eql x y))
@@ -508,16 +505,13 @@ length and have identical components. Other arrays must be EQ to be EQUAL."
         (t nil)))
 
 #-sb-show ;; I don't know why these tests crash with #+sb-show
-(let ((test-cases `(($0.0 $-0.0 t)
-                    ($0.0 $1.0 nil)
-                    ;; There is no cross-compiler #C reader macro.
-                    ;; SB-XC:COMPLEX does not want uncanonical input, i.e. imagpart
-                    ;; of rational 0 which downgrades the result to just an integer.
-                    (1 ,(complex $1.0 $0.0) t)
-                    (,(complex 0 1) ,(complex $0.0 $1.0) t)
+(let ((test-cases '((0.0 -0.0 t)
+                    (0.0 1.0 nil)
+                    (#c(1 0) #c(1.0 0.0) t)
+                    (#c(0 1) #c(0.0 1.0) t)
                     ;; 11/10 is unequal to real 1.1 due to roundoff error.
                     ;; COMPLEX here is a red herring
-                    (,(complex $1.1 $0.0) 11/10 nil)
+                    (#c(1.1 0.0) #c(11/10 0) nil)
                     ("Hello" "hello" t)
                     ("Hello" #(#\h #\E #\l #\l #\o) t)
                     ("Hello" "goodbye" nil))))
