@@ -58,7 +58,8 @@
 ;;; Since WITH-PINNED-OBJECT costs nothing on conservative gencgc,
 ;;; that's what I'm going with.
 ;;;
-(defun %instance-sxhash (instance)
+(declaim (inline %instance-sxhash))
+(defun %instance-sxhash (instance header-word)
   ;; LAYOUT must not acquire an extra slot for the stable hash,
   ;; because the bitmap length is derived from the instance length.
   ;; It would probably be simple to eliminate this as a special case
@@ -68,28 +69,31 @@
     ;; This might be wrong if the clos-hash was clobbered to 0
     (return-from %instance-sxhash (layout-clos-hash instance)))
   ;; Non-simple cases: no hash slot, and either unhashed or hashed-not-moved.
-  (let* ((header-word (instance-header-word instance))
-         (addr (with-pinned-objects (instance)
-                 ;; First we have to indicate that a hash was taken from the address
-                 ;; if not already so marked.
-                 (unless (logbitp sb-vm:stable-hash-required-flag header-word)
-                   #-sb-thread (setf (sap-ref-word (int-sap (get-lisp-obj-address instance))
-                                                   (- sb-vm:instance-pointer-lowtag))
-                                     (logior (ash 1 sb-vm:stable-hash-required-flag)
-                                             header-word))
-                   #+sb-thread (%primitive sb-vm::set-instance-hashed instance))
-                 (get-lisp-obj-address instance))))
+  (let* ((addr (sb-c::if-vop-existsp (:named sb-vm::set-instance-hashed-return-address)
+                 (if (logbitp sb-vm:stable-hash-required-flag header-word)
+                     (get-lisp-obj-address instance)
+                     (%primitive sb-vm::set-instance-hashed-return-address instance))
+                 (with-pinned-objects (instance)
+                   ;; First we have to indicate that a hash was taken from the address
+                   ;; if not already so marked.
+                   (unless (logbitp sb-vm:stable-hash-required-flag header-word)
+                     #-sb-thread (setf (sap-ref-word (int-sap (get-lisp-obj-address instance))
+                                                     (- sb-vm:instance-pointer-lowtag))
+                                       (logior (ash 1 sb-vm:stable-hash-required-flag)
+                                               header-word))
+                     #+sb-thread (%primitive sb-vm::set-instance-hashed instance))
+                   (get-lisp-obj-address instance)))))
     ;; perturb the address
     (murmur-hash-word/+fixnum addr)))
 
 (declaim (inline instance-sxhash))
 (defun instance-sxhash (instance)
-  (if (logbitp sb-vm:hash-slot-present-flag
-               (instance-header-word (truly-the instance instance)))
-      ;; easy case: 1 word beyond the apparent length is a word added
-      ;; by GC (which may have resized the object, but we don't need to know).
-      (truly-the hash-code (%instance-ref instance (%instance-length instance)))
-      (%instance-sxhash instance)))
+  (let ((header-word (instance-header-word (truly-the instance instance))))
+    (if (logbitp sb-vm:hash-slot-present-flag header-word)
+        ;; easy case: 1 word beyond the apparent length is a word added
+        ;; by GC (which may have resized the object, but we don't need to know).
+        (truly-the hash-code (%instance-ref instance (%instance-length instance)))
+        (%instance-sxhash instance header-word))))
 
 ;;; Return a pseudorandom number that was assigned on allocation.
 ;;; FIN is a STANDARD-FUNCALLABLE-INSTANCE but we don't care to type-check it.
@@ -399,80 +403,74 @@
                           do (mixf result (number-psxhash (funcall getter data i)))))))))
              result))
        (structure-object-psxhash (key depthoid)
-       ;; Compute a PSXHASH for KEY. Salient points:
-       ;; * It's not enough to use the bitmap to figure out how to mix in raw slots.
-       ;;   The floating-point types all need special treatment. And we want to avoid
-       ;;   consing, so we can't very well call PSXHASH.
-       ;; * Even though PSXHASH requires that numerically equal numbers have the same
-       ;;   hash e.g. 12 and 12d0 and #c(12d0 0d0) all hash the same, structures can
-       ;;   weaken that restriction: instances are EQUAL only if they are of the same
-       ;;   type and slot-for-slot EQUAL. So a float in a raw slot can't be EQUAL
-       ;;   to a word in a different raw slot. In fact we don't even require that
-       ;;   SINGLE- and DOUBLE-float hash the same for a given numerical value,
-       ;;   because a raw slot can't hold either/or. But -0 and +0 must hash the same.
-       (declare (type structure-object key))
-       (declare (type (integer 0 #.+max-hash-depthoid+) depthoid))
-       (macrolet ((rsd-index+1 (dsd)
-                    ;; Return 0 if the DSD is not raw, otherwise 1+ the index into
-                    ;; *RAW-SLOT-DATA*. This is exactly the low 3 bits of DSD-BITS.
-                    `(truly-the (mod ,(1+ (length sb-kernel::*raw-slot-data*)))
-                                (ldb (byte 3 0) (sb-kernel::dsd-bits ,dsd))))
-                  (raw-cases ()
-                    (flet ((1+index-of (type)
-                             (1+ (position type sb-kernel::*raw-slot-data*
-                                           :key #'sb-kernel::raw-slot-data-raw-type)))
-                           (mix-float (val zero)
-                             `(let ((x ,val))
-                                (mixf result (sxhash (if (= x ,zero) ,zero x))))))
-                      ;; This compiles to a jump table if supported
-                      `(case rsd-index+1
-                        ((,(1+index-of 'word) ,(1+index-of 'sb-vm:signed-word))
-                         ;; Access as unsigned. +X and -X hash differently because
-                         ;; of 2's complement, so disregarding the sign bit is fine.
-                         (mixf result (logand (%raw-instance-ref/word key i)
-                                              most-positive-fixnum)))
-                        (,(1+index-of 'single-float)
-                         ,(mix-float '(%raw-instance-ref/single key i) $0f0))
-                        (,(1+index-of 'double-float)
-                         ,(mix-float '(%raw-instance-ref/double key i) $0d0))
-                        (,(1+index-of 'sb-kernel:complex-single-float)
-                         (let ((cplx (%raw-instance-ref/complex-single key i)))
-                           ,(mix-float '(realpart cplx) $0f0)
-                           ,(mix-float '(imagpart cplx) $0f0)))
-                        (,(1+index-of 'sb-kernel:complex-double-float)
-                         (let ((cplx (%raw-instance-ref/complex-double key i)))
-                           ,(mix-float '(realpart cplx) $0d0)
-                           ,(mix-float '(imagpart cplx) $0d0)))))))
-         (let* ((layout (%instance-layout key))
-                (result (layout-clos-hash layout)))
-           (declare (type fixnum result))
-           (when (plusp depthoid)
-             (let ((max-iterations depthoid)
-                   (depthoid (1- depthoid))
-                   (dd (layout-dd layout)))
-               (declare (index max-iterations))
-               (if (/= (sb-kernel::dd-bitmap dd) +layout-all-tagged+)
-                   (let ((slots (dd-slots dd)))
-                     (loop (unless slots (return))
-                           (let* ((slot (pop slots))
-                                  (rsd-index+1 (rsd-index+1 slot))
-                                  (i (dsd-index slot)))
-                             (cond ((= rsd-index+1 0) ; non-raw
-                                    (mixf result (%psxhash (%instance-ref key i) depthoid))
-                                    (if (zerop (decf max-iterations)) (return)))
-                                   (t
-                                    ;; Don't decrement MAX-ITERATIONS.
-                                    ;; These can't cause unbounded work.
-                                    (raw-cases))))))
-                   (let ((len (%instance-length key))
-                         ;; Don't mix in LAYOUT (if it takes a slot) because it was the seed value.
-                         (i sb-vm:instance-data-start))
-                     (declare (index i))
-                     (loop (when (>= i len) (return))
-                           (mixf result (%psxhash (%instance-ref key i) depthoid))
-                           (incf i)
-                           (if (zerop (decf max-iterations)) (return)))))))
-           result)))
+         ;; Compute a PSXHASH for KEY. Salient points:
+         ;; * It's not enough to use the bitmap to figure out how to mix in raw slots.
+         ;;   The floating-point types all need special treatment. And we want to avoid
+         ;;   consing, so we can't very well call PSXHASH.
+         ;; * Even though PSXHASH requires that numerically equal numbers have the same
+         ;;   hash e.g. 12 and 12d0 and #c(12d0 0d0) all hash the same, structures can
+         ;;   weaken that restriction: instances are EQUAL only if they are of the same
+         ;;   type and slot-for-slot EQUAL. So a float in a raw slot can't be EQUAL
+         ;;   to a word in a different raw slot. In fact we don't even require that
+         ;;   SINGLE- and DOUBLE-float hash the same for a given numerical value,
+         ;;   because a raw slot can't hold either/or. But -0 and +0 must hash the same.
+         (declare (type structure-object key))
+         (declare (type (integer 0 #.+max-hash-depthoid+) depthoid))
+         (macrolet ((rsd-index+1 (dsd)
+                      ;; Return 0 if the DSD is not raw, otherwise 1+ the index into
+                      ;; *RAW-SLOT-DATA*. This is exactly the low 3 bits of DSD-BITS.
+                      `(truly-the (mod ,(1+ (length sb-kernel::*raw-slot-data*)))
+                         (ldb (byte 3 0) (sb-kernel::dsd-bits ,dsd))))
+                    (raw-cases ()
+                      (flet ((1+index-of (type)
+                               (1+ (position type sb-kernel::*raw-slot-data*
+                                             :key #'sb-kernel::raw-slot-data-raw-type)))
+                             (mix-float (val zero)
+                               `(let ((x ,val))
+                                  (mixf result (sxhash (if (= x ,zero) ,zero x))))))
+                        ;; This compiles to a jump table if supported
+                        `(case rsd-index+1
+                           ((,(1+index-of 'word) ,(1+index-of 'sb-vm:signed-word))
+                            ;; Access as unsigned. +X and -X hash differently because
+                            ;; of 2's complement, so disregarding the sign bit is fine.
+                            (mixf result (logand (%raw-instance-ref/word key i)
+                                                 most-positive-fixnum)))
+                           (,(1+index-of 'single-float)
+                            ,(mix-float '(%raw-instance-ref/single key i) 0f0))
+                           (,(1+index-of 'double-float)
+                            ,(mix-float '(%raw-instance-ref/double key i) 0d0))
+                           (,(1+index-of 'sb-kernel:complex-single-float)
+                            (let ((cplx (%raw-instance-ref/complex-single key i)))
+                              ,(mix-float '(realpart cplx) 0f0)
+                              ,(mix-float '(imagpart cplx) 0f0)))
+                           (,(1+index-of 'sb-kernel:complex-double-float)
+                            (let ((cplx (%raw-instance-ref/complex-double key i)))
+                              ,(mix-float '(realpart cplx) 0d0)
+                              ,(mix-float '(imagpart cplx) 0d0)))))))
+           (let* ((layout (%instance-layout key))
+                  (result (layout-clos-hash layout)))
+             (declare (type fixnum result))
+             (when (plusp depthoid)
+               (let ((depthoid (1- depthoid))
+                     (dd (layout-dd layout)))
+                 (if (/= (sb-kernel::dd-bitmap dd) +layout-all-tagged+)
+                     (let ((slots (dd-slots dd)))
+                       (loop (unless slots (return))
+                             (let* ((slot (pop slots))
+                                    (rsd-index+1 (rsd-index+1 slot))
+                                    (i (dsd-index slot)))
+                               (cond ((= rsd-index+1 0) ; non-raw
+                                      (mixf result (%psxhash (%instance-ref key i) depthoid)))
+                                     (t
+                                      (raw-cases))))))
+                     (let ((len (%instance-length key))
+                           ;; Don't mix in LAYOUT (if it takes a slot) because it was the seed value.
+                           (i sb-vm:instance-data-start))
+                       (declare (index i))
+                       (loop (when (>= i len) (return))
+                             (mixf result (%psxhash (%instance-ref key i) depthoid))
+                             (incf i))))))
+             result)))
        (sfloat-psxhash (key)
          (declare (single-float key))
          (hash-float single-float key))
@@ -572,4 +570,3 @@
 
 ;;; Not needed post-build
 (clear-info :function :inlining-data '%sxhash-simple-substring)
-
