@@ -67,14 +67,17 @@
         (abort-ir1-transform)
         expansion)))
 
-(progn
-  (defun type-other-pointer-p (type)
-    (csubtypep type (specifier-type '(not #1=(or fixnum #+64-bit single-float
-                                                 list function instance character
-                                                 extended-sequence)))))
+(sb-xc:deftype other-pointer ()
+  '(or array
+    (and number (not (or fixnum #+64-bit single-float)))
+    fdefn (and symbol (not null))
+    weak-pointer system-area-pointer code-component))
 
-  (defun type-not-other-pointer-p (type)
-    (csubtypep type (specifier-type '#1#))))
+(defun type-other-pointer-p (type)
+  (csubtypep type (specifier-type 'other-pointer)))
+
+(defun type-not-other-pointer-p (type)
+  (csubtypep type (specifier-type '(not other-pointer))))
 
 ;;; If the lvar OBJECT definitely is or isn't of the specified
 ;;; type, then return T or NIL as appropriate. Otherwise quietly
@@ -183,7 +186,7 @@
                              (cond ((and (type-not-other-pointer-p diff)
                                          (type-other-pointer-p type))
                                     `(%other-pointer-p object))
-                                   ((and (type-other-pointer-p type)
+                                   ((and (type-other-pointer-p diff)
                                          (type-not-other-pointer-p type))
                                     `(not (%other-pointer-p object)))))))))
               (t
@@ -218,14 +221,15 @@
     `(or (classoid-cell-classoid ',cell)
          (error "Class not yet defined: ~S" name))))
 
-(defoptimizer (%typep-wrapper constraint-propagate-if)
-    ((test-value variable type) node)
+(defoptimizer (%typep-wrapper constraint-propagate-if) ((test-value variable type) node)
   (aver (constant-lvar-p type))
-  (let ((type (lvar-value type)))
-    (values variable (if (ctype-p type)
-                         type
-                         (handler-case (careful-specifier-type type)
-                           (t () nil))))))
+  (let* ((type (lvar-value type))
+         (ctype (if (ctype-p type)
+                    type
+                    (handler-case (careful-specifier-type type)
+                      (t () nil)))))
+    (if (and ctype (type-for-constraints-p ctype))
+        (values variable ctype))))
 
 (deftransform %typep-wrapper ((test-value variable type) * * :node node)
   (aver (constant-lvar-p type))
@@ -342,22 +346,19 @@
 
 ;;; Return a form that tests the variable N-OBJECT for being in the
 ;;; binds specified by TYPE. BASE is the name of the base type, for
-;;; declaration. We make SAFETY locally 0 to inhibit any checking of
-;;; this assertion.
+;;; declaration.
 (defun transform-numeric-bound-test (n-object type base)
   (declare (type numeric-type type))
   (let ((low (numeric-type-low type))
         (high (numeric-type-high type)))
-    `(locally
-       (declare (optimize (safety 0)))
-       (and ,@(when low
-                (if (consp low)
-                    `((> (truly-the ,base ,n-object) ,(car low)))
-                    `((>= (truly-the ,base ,n-object) ,low))))
-            ,@(when high
-                (if (consp high)
-                    `((< (truly-the ,base ,n-object) ,(car high)))
-                    `((<= (truly-the ,base ,n-object) ,high))))))))
+    `(and ,@(when low
+              (if (consp low)
+                  `((> (truly-the ,base ,n-object) ,(car low)))
+                  `((>= (truly-the ,base ,n-object) ,low))))
+          ,@(when high
+              (if (consp high)
+                  `((< (truly-the ,base ,n-object) ,(car high)))
+                  `((<= (truly-the ,base ,n-object) ,high)))))))
 
 ;;; Do source transformation of a test of a known numeric type. We can
 ;;; assume that the type doesn't have a corresponding predicate, since
@@ -390,32 +391,7 @@
          (high (numeric-type-high type)))
     (ecase (numeric-type-complexp type)
       (:real
-       (cond #-arm64
-             ((and (eql (numeric-type-class type) 'integer)
-                   (and (fixnump low)
-                        (fixnump high)
-                        #+(or x86 x86-64 arm arm64)
-                        (/= low 0)
-                        (< (- high low) 2)))
-              ;; The fixnum-mod-p case is worse than just EQ testing with
-              ;; only 2 values in the range. (INTEGER 1 2) would have become
-              ;;   (and (not (eq x 0)) (fixnump x) (not (> x 2))).
-              ;; If exactly 1 value, it should have been picked off by TYPE-SINGLETON-P
-              ;; in %SOURCE-TRANSFORM-TYPEP, but even if it wasn't,
-              ;; the OR will drop out due to constraint propagation.
-              `(or (eq ,object ,low) (eq ,object ,high)))
-             ((and (vop-existsp :translate fixnum-mod-p)
-                   (eql (numeric-type-class type) 'integer)
-                   (or (eql low 0)
-                       (and (eql low 1)
-                            (not (eql high most-positive-fixnum))))
-                   (fixnump high))
-              (let ((mod-p `(fixnum-mod-p ,object ,high)))
-                (if (eql low 1)
-                    `(and (not (eq ,object 0))
-                          ,mod-p)
-                    mod-p)))
-             ((and (vop-existsp :translate check-range<=)
+       (cond ((and (vop-existsp :translate check-range<=)
                    (eql (numeric-type-class type) 'integer)
                    (fixnump low)
                    (fixnump high))
@@ -423,21 +399,25 @@
              ((type= type (specifier-type '(or word sb-vm:signed-word)))
               `(or (typep ,object 'sb-vm:signed-word)
                    (typep ,object 'word)))
+             ((and (vop-existsp :translate unsigned-byte-x-p)
+                   (eql (numeric-type-class type) 'integer)
+                   (eql low 0)
+                   (integerp high)
+                   (= (logcount (1+ high)) 1)
+                   (zerop (rem (integer-length high) sb-vm:n-word-bits)))
+              `(unsigned-byte-x-p ,object ,(integer-length high)))
              (t
               `(and (typep ,object ',base)
                     ,(transform-numeric-bound-test object type base)))))
       (:complex
-       `(and (complexp ,object)
-             ,(once-only ((n-real `(realpart (truly-the complex ,object)))
-                          (n-imag `(imagpart (truly-the complex ,object))))
-                `(progn
-                   ,n-imag ; ignorable
-                   (and (typep ,n-real ',base)
-                        ,@(when (eq class 'integer)
-                            `((typep ,n-imag ',base)))
-                        ,(transform-numeric-bound-test n-real type base)
-                        ,(transform-numeric-bound-test n-imag type
-                                                       base)))))))))
+       (let ((part-type (second (type-specifier type))))
+         `(and (typep ,object '(complex ,(case base
+                                           ((double-float single-float rational) base)
+                                           (t (if (eq class 'integer)
+                                                  'rational
+                                                  '*)))))
+               (typep (realpart ,object) ',part-type)
+               (typep (imagpart ,object) ',part-type)))))))
 
 ;;; Do the source transformation for a test of a hairy type.
 ;;; SATISFIES is converted into the obvious. Otherwise, we convert
@@ -454,20 +434,6 @@
            ;; as soon as any unknown is present.
            `(classoid-cell-typep ,(find-classoid-cell spec :create t) ,object))
           ((unknown-type-p type)
-           #+(and sb-xc-host (not sb-devel))
-           (warn "can't open-code test of unknown type ~S"
-                 (type-specifier type))
-           ;; This is not a policy-based decision to notify here,
-           ;; because it is _ALWAYS_ questionable style imho to refer to unknown types.
-           ;; Unfortunately, people love to suppress COMPILER-NOTE because SBCL produces
-           ;; far too many of those for low-level things like untagged-SAP-to-tagged-SAP.
-           ;; So we could opt to STYLE-WARN, which is, in this case, perhaps more severe
-           ;; than we'd like?
-           ;; I guess we're just going to have to say that if you've muffled too may
-           ;; kinds of NOTEs, that's on you.
-           #-sb-xc-host (compiler-notify 'unknown-typep-note
-                                         :format-control "can't open-code test of unknown type ~S"
-                                         :format-arguments (list (type-specifier type)))
            `(let ((object ,object)
                   (cache (load-time-value (cons #'sb-kernel::cached-typep ',spec)
                                           t)))
@@ -597,15 +563,47 @@
              `((typep ,object '(or ,@(mapcar #'type-specifier other))))))))))
 
 (defun source-transform-union-numeric-typep (object types)
-  (cond ((not (every #'numeric-type-p types))
+  (cond ((and (= (length types) 2)
+              ;; Transform (or (double-float * (0d0)) (eql -0d0))
+              (destructuring-bind (a b) types
+                (multiple-value-bind (member numeric) (cond ((member-type-p a)
+                                                             (values a b))
+                                                            ((member-type-p b)
+                                                             (values b a)))
+                  (let ((double))
+                    (and (numeric-type-p numeric)
+                         (= (member-type-size member) 1)
+                         (or (setf double (sb-kernel::member-type-member-p -0d0 member))
+                             (sb-kernel::member-type-member-p -0f0 member))
+                         (let ((low (numeric-type-low numeric))
+                               (high (numeric-type-high numeric))
+                               (type (if double
+                                         'double-float
+                                         'single-float)))
+                           (when (and (eq (numeric-type-class numeric) 'float)
+                                      (eq (numeric-type-complexp numeric) :real)
+                                      (equal high (if double
+                                                      '(0d0)
+                                                      '(0f0))))
+                             `(and ,(if double
+                                        `(double-float-p ,object)
+                                        `(single-float-p ,object))
+                                   ,(if low
+                                        `(and (float-sign-bit-set-p (truly-the ,type ,object))
+                                              (,@(if (consp low)
+                                                     `(< ,(car low))
+                                                     `(<= ,low))
+                                               (truly-the ,type ,object)))
+                                        `(float-sign-bit-set-p (truly-the ,type ,object))))))))))))
+        ((not (every #'numeric-type-p types))
          nil)
         ((and (= (length types) 2)
               ;; (and subtype-of-integer (not (eql x)))
               ;; don't test a range.
               ;; (and subtype-of-integer (not (integer x y)))
               (destructuring-bind (b a) types
-                (and (eq (numeric-type-class a) 'integer)
-                     (eq (numeric-type-class b) 'integer)
+                (and (integer-type-p a)
+                     (integer-type-p b)
                      (flet ((check (a b)
                               (let* ((a-hi (numeric-type-high a))
                                      (a-lo (numeric-type-low a))
@@ -636,8 +634,8 @@
               ;; (or (integer * fixnum-x) (integer fixnum-y))
               ;; only check for bignump and not its value.
               (destructuring-bind (b a) types
-                (and (eq (numeric-type-class a) 'integer)
-                     (eq (numeric-type-class b) 'integer)
+                (and (integer-type-p a)
+                     (integer-type-p b)
                      (flet ((check (a b)
                               (let* ((a-hi (numeric-type-high a))
                                      (a-lo (numeric-type-low a))
@@ -965,7 +963,8 @@
                                                    (let ((widetag (%other-pointer-widetag data)))
                                                      (if (eq widetag ,typecode)
                                                          (return t)
-                                                         (unless (>= widetag sb-vm:complex-vector-widetag)
+                                                         (unless (or (eq widetag sb-vm:simple-array-widetag)
+                                                                     (>= widetag sb-vm:complex-vector-widetag))
                                                            (return nil)))))))))))
                               t
                               t)
@@ -980,7 +979,9 @@
                                                    (let ((widetag (%other-pointer-widetag data)))
                                                      (if (eq widetag ,typecode)
                                                          (return t)
-                                                         (unless (>= widetag sb-vm:complex-vector-widetag)
+                                                         (unless (or
+                                                                  (eq widetag sb-vm:simple-array-widetag)
+                                                                  (>= widetag sb-vm:complex-vector-widetag))
                                                            (return nil)))))))))))
                               t))))))))))
 
@@ -1154,11 +1155,6 @@
                 `(and (%instancep object)
                       (logtest (,get-flags (%instance-layout object)) ,flag)))))
 
-          ;; TODO: remove after April 2021 release.
-          ((eq name 'sb-kernel::random-class)
-           (style-warn "~S should not appear in a TYPEP test" name)
-           nil)
-
           ;; Next easiest: Sealed and no subtypes. Typically for DEFSTRUCT only.
           ;; Even if you don't seal a DEFCLASS, we're allowed to assume that things
           ;; won't change, as per CLHS 3.2.2.3 on Semantic Constraints:
@@ -1473,7 +1469,7 @@
                     (error "~a is not a subtype of VECTOR." type)))))
     (simplify type)))
 
-(defun strip-array-dimensions-and-complexity (type)
+(defun strip-array-dimensions-and-complexity (type &optional simple)
   (labels ((process-compound-type (types)
              (let (array-types)
                (dolist (type types)
@@ -1489,7 +1485,9 @@
                            dim
                            (make-list (length dim)
                                       :initial-element '*))
-                       :complexp :maybe
+                       :complexp (if simple
+                                     nil
+                                     :maybe)
                        :element-type (array-type-element-type type)
                        :specialized-element-type (array-type-specialized-element-type type))))
                    ((union-type-p type)
@@ -1641,7 +1639,12 @@
                            `(make-array ,dimension ,@specialization :initial-contents x))))))))
       ((type= tspec (specifier-type 'list))
        `(coerce-to-list x))
-      ((csubtypep tspec (specifier-type 'function))
+      ((csubtypep tspec (specifier-type 'extended-sequence))
+       (let ((class (and (symbolp tval) (find-class tval nil))))
+         (if (null class)
+             (give-up-ir1-transform)
+             `(coerce-to-extended-sequence x (load-time-value (find-class ',tval) t)))))
+      ((type= tspec (specifier-type 'function))
        (if (csubtypep (lvar-type x) (specifier-type 'symbol))
            `(coerce-symbol-to-fun x)
            ;; if X can later be derived as FUNCTION then we don't want
@@ -1649,6 +1652,12 @@
            ;; that can undo that and see that it's really (IDENTITY X).
            (progn (delay-ir1-transform node :constraint)
                   `(coerce-to-fun x))))
+      ((multiple-value-bind (p really)
+           (csubtypep tspec
+                      (specifier-type '(or sequence character complex float function)))
+         (and really
+              (not p)))
+       `(the* (,tspec :context coerce-context) x))
       (t
        (give-up-ir1-transform
         "~@<open coding coercion to ~S not implemented.~:@>"
@@ -1658,22 +1667,34 @@
   ((value) (sb-vm:signed-word) * :important nil)
   `(>= value 0))
 
+(when-vop-existsp (:translate unsigned-byte-x-p)
+  (deftransform unsigned-byte-x-p
+      ((value x) (t t) * :important nil :node node)
+    (ir1-transform-type-predicate value (specifier-type `(unsigned-byte ,(lvar-value x))) node))
+
+  (deftransform unsigned-byte-x-p
+      ((value x) ((integer * #.most-positive-word) t) * :important nil)
+    `(#+64-bit unsigned-byte-64-p #-64-bit unsigned-byte-32-p x)))
+
 (deftransform %other-pointer-p ((object))
-  (let ((this-type
-          (specifier-type '(or fixnum
-                            #+64-bit single-float
-                            function
-                            list
-                            instance
-                            character))))
-    (cond ((not (types-equal-or-intersect this-type (lvar-type object))))
-          ((csubtypep (lvar-type object) this-type)
+  (let ((type (lvar-type object)))
+    (cond ((not (types-equal-or-intersect type (specifier-type 'other-pointer)))
            nil)
+          ((or (csubtypep type (specifier-type 'other-pointer))
+               ;; It doesn't negate to this type, so check both
+               (csubtypep type (specifier-type '(not (or fixnum #+64-bit single-float
+                                                                list function instance character)))))
+           t)
           ((give-up-ir1-transform)))))
 
 ;;; BIGNUMP is simpler than INTEGERP, so if we can rule out FIXNUM then ...
 (deftransform integerp ((x) ((not fixnum)) * :important nil) '(bignump x))
 
+(deftransform structure-typep ((object type) (t t) * :node node)
+  (if (types-equal-or-intersect (lvar-type object) (specifier-type 'instance))
+      (give-up-ir1-transform)
+      nil))
+
 (deftransform structure-typep ((object type) (t (constant-arg t)))
   (let* ((layout (lvar-value type))
          (type (case layout
@@ -1684,9 +1705,14 @@
                   (layout-classoid layout))))
          (diff (type-difference (lvar-type object) type))
          (pred (backend-type-predicate diff)))
-    (if pred
-        `(not (,pred object))
-        (give-up-ir1-transform))))
+    (cond ((not (types-equal-or-intersect (lvar-type object) type))
+           nil)
+          ((csubtypep (lvar-type object) type)
+           t)
+          (pred
+           `(not (,pred object)))
+          (t
+           (give-up-ir1-transform)))))
 
 (deftransform classoid-cell-typep ((cell object) ((constant-arg t) t))
   (let* ((type (specifier-type (classoid-cell-name (lvar-value cell))))
@@ -1696,26 +1722,6 @@
         `(not (,pred object))
         (give-up-ir1-transform))))
 
-(when-vop-existsp (:translate fixnum-mod-p)
-  (deftransform fixnum-mod-p ((x mod) (t (constant-arg fixnum)) * :important nil)
-    (let* ((type (lvar-type x))
-           (mod (lvar-value mod))
-           (intersect (type-intersection type (specifier-type 'fixnum)))
-           (mod-type (specifier-type `(mod ,(1+ mod)))))
-      (cond ((csubtypep intersect mod-type)
-             `(fixnump x))
-            ((and (csubtypep type (specifier-type 'fixnum))
-                  (csubtypep (type-intersection type (specifier-type 'unsigned-byte))
-                             mod-type))
-             `(>= x 0))
-            ((let ((int (type-approximate-interval intersect)))
-               (when int
-                 (let ((power-of-two (1- (ash 1 (integer-length (interval-high int))))))
-                   (when (< 0 power-of-two mod)
-                     `(fixnum-mod-p x ,power-of-two))))))
-            (t
-             (give-up-ir1-transform))))))
-
 (when-vop-existsp (:translate signed-byte-8-p)
   (macrolet ((def (bits)
                `(deftransform ,(symbolicate "SIGNED-BYTE-" (princ-to-string bits) "-P")
