@@ -13,7 +13,7 @@
  * files for more information.
  */
 
-#include "sbcl.h"
+#include "genesis/sbcl.h"
 
 #include <stdio.h>
 #include <string.h>
@@ -56,7 +56,6 @@
 #include "gc.h"
 #include "validate.h"
 #include "core.h"
-#include "save.h"
 #include "lispregs.h"
 #include "thread.h"
 
@@ -386,19 +385,7 @@ static int is_memsize_arg(char *argv[], int argi, int argc, int *merge_core_page
         if ((argi+1) >= argc) lose("missing argument for --dynamic-space-size");
         dynamic_space_size = parse_size_arg(argv[argi+1],
                                             "--dynamic-space-size");
-#ifdef MAX_DYNAMIC_SPACE_END
-        if (!((DYNAMIC_SPACE_START <
-                       DYNAMIC_SPACE_START+dynamic_space_size) &&
-                      (DYNAMIC_SPACE_START+dynamic_space_size <=
-                       MAX_DYNAMIC_SPACE_END))) {
-            char* suffix = "";
-            char* size = argv[argi-1];
-            if (!strchr(size, 'B') && !strchr(size, 'b')) suffix = " [MB]";
-            lose("--dynamic-space-size argument %s%s is too large, max %lu KB",
-                 size, suffix, (MAX_DYNAMIC_SPACE_END-DYNAMIC_SPACE_START) / 1024);
-        }
-#endif
-        return 2;
+        return 2; // return number of elements of argv[] consumed
     }
     if (!strcmp(arg, "--control-stack-size")) {
         if ((argi+1) >= argc) lose("missing argument for --control-stack-size");
@@ -463,27 +450,31 @@ parse_argv(struct memsize_options memsize_options,
         dynamic_space_size = memsize_options.dynamic_space_size;
         thread_control_stack_size = memsize_options.thread_control_stack_size;
         dynamic_values_bytes = memsize_options.thread_tls_bytes;
-#ifndef LISP_FEATURE_WIN32
-        sbcl_argv = successful_malloc((argc + 1) * sizeof(char *));
-        sbcl_argv[0] = argv[0];
         int stop_parsing = 0; // have we seen '--'
         int output_index = 1;
+
+#ifndef LISP_FEATURE_WIN32
+        sbcl_argv = successful_malloc((argc + 1) * sizeof(char *));
+        char **argv_source = argv;
+#else
+        int wargc;
+        wchar_t **argv_source;
+        argv_source = CommandLineToArgvW(GetCommandLineW(), &wargc);
+        sbcl_argv = successful_malloc((wargc + 1) * sizeof(wchar_t *));
+#endif
+        sbcl_argv[0] = argv_source[0];
+
         while (argi < argc) {
             if (stop_parsing) // just copy it over
-                sbcl_argv[output_index++] = argv[argi++];
+                sbcl_argv[output_index++] = argv_source[argi++];
             else if (!strcmp(argv[argi], "--")) // keep it, but parse nothing else
-                sbcl_argv[output_index++] = argv[argi++], stop_parsing = 1;
+                sbcl_argv[output_index++] = argv_source[argi++], stop_parsing = 1;
             else if ((n_consumed = is_memsize_arg(argv, argi, argc, &merge_core_pages)))
                 argi += n_consumed; // eat it
             else // default action - copy it
-                sbcl_argv[output_index++] = argv[argi++];
+                sbcl_argv[output_index++] = argv_source[argi++];
         }
         sbcl_argv[output_index] = 0;
-#else
-        int wargc;
-        sbcl_argv = CommandLineToArgvW(GetCommandLineW(), &wargc);
-        // Somebody who wishes this to work for #+win32 should feel free to do the same...
-#endif
     } else {
         bool end_runtime_options = 0;
         /* Parse our any of the command-line options that we handle from C,
@@ -613,6 +604,12 @@ initialize_lisp(int argc, char *argv[], char *envp[])
     struct lisp_exception_frame exception_frame;
 #endif
 #ifdef LISP_FEATURE_UNIX
+#ifdef LISP_FEATURE_AVOID_CLOCK_GETTIME
+    struct timeval tv;
+    gettimeofday(&tv, NULL);
+    lisp_init_time.tv_sec = tv.tv_sec;
+    lisp_init_time.tv_nsec = tv.tv_usec * 1000;
+#else
     clock_gettime(
 #ifdef LISP_FEATURE_LINUX
         CLOCK_MONOTONIC_COARSE
@@ -620,6 +617,7 @@ initialize_lisp(int argc, char *argv[], char *envp[])
         CLOCK_MONOTONIC
 #endif
         , &lisp_init_time);
+#endif
 #endif
 
     /* the name of the core file we're to execute. Note that this is
@@ -636,30 +634,46 @@ initialize_lisp(int argc, char *argv[], char *envp[])
 
     interrupt_init();
 #ifdef LISP_FEATURE_UNIX
-    /* Not sure why anyone sends signals to this process so early.
-     * But win32 models the signal mask as part of 'struct thread'
-     * which doesn't exist yet, so don't do this */
-    block_blockable_signals(0);
+    /* Use SETMASK instead of BLOCK to clear the inhereted sigmask. */
+    thread_sigmask(SIG_SETMASK, &blockable_sigset, 0);
 #endif
 
     /* Check early to see if this executable has an embedded core,
      * which also populates runtime_options if the core has runtime
      * options */
-    if (!(sbcl_runtime = os_get_runtime_executable_path()))
-        sbcl_runtime = search_for_executable(argv[0]);
-
-    if (!(sbcl_runtime_home = dir_name(argv[0])))
-      if (!(sbcl_runtime_home = dir_name(sbcl_runtime)))
-        sbcl_runtime_home = libpath;
-
+    sbcl_runtime = os_get_runtime_executable_path();
     if (sbcl_runtime) {
         os_vm_offset_t offset = search_for_embedded_core(sbcl_runtime, &memsize_options);
         if (offset != -1) {
-            embedded_core_offset = offset;
             core = sbcl_runtime;
+            embedded_core_offset = offset;
         }
     }
 
+    if (!core) {
+        char *exe_path = search_for_executable(argv[0]);
+        if (exe_path && !(sbcl_runtime && strcmp(sbcl_runtime, exe_path) == 0)) {
+            os_vm_offset_t offset = search_for_embedded_core(exe_path, &memsize_options);
+            if (offset != -1) {
+                free(core);
+                sbcl_runtime = exe_path;
+                core = exe_path;
+                embedded_core_offset = offset;
+            } else {
+                if (!sbcl_runtime)
+                    sbcl_runtime = exe_path;
+                else
+                    free(exe_path);
+            }
+        } else {
+            free(exe_path);
+        }
+    }
+
+    if (!(sbcl_runtime_home = dir_name(argv[0])))
+        if (!(sbcl_runtime_home = dir_name(sbcl_runtime)))
+            sbcl_runtime_home = libpath;
+
     struct cmdline_options options = parse_argv(memsize_options, argc, argv, envp, core);
 
     /* Align down to multiple of page_table page size, and to the appropriate
@@ -683,7 +697,8 @@ initialize_lisp(int argc, char *argv[], char *envp[])
     if (!core && !(core = search_for_core())) {
       /* Try resolving symlinks */
       if (sbcl_runtime) {
-        free(sbcl_runtime_home);
+        if (sbcl_runtime_home != libpath)
+            free(sbcl_runtime_home);
         char* real = sb_realpath(sbcl_runtime);
         if (!real)
           goto lose;
@@ -756,6 +771,12 @@ initialize_lisp(int argc, char *argv[], char *envp[])
     write_protect_immobile_space();
 #endif
 
+    /* Requires STATIC_SPACE_START to have been calculated. */
+#if defined(LISP_FEATURE_SB_SAFEPOINT)
+    extern void safepoint_init(void);
+    safepoint_init();
+#endif
+
     arch_install_interrupt_handlers();
 #ifndef LISP_FEATURE_WIN32
     os_install_interrupt_handlers();
