@@ -32,7 +32,7 @@
 #include <sys/param.h>
 #include <sys/file.h>
 #include <io.h>
-#include "sbcl.h"
+#include "genesis/sbcl.h"
 #include "os.h"
 #include "arch.h"
 #include "globals.h"
@@ -40,9 +40,7 @@
 #include "interr.h"
 #include "lispregs.h"
 #include "runtime.h"
-#include "alloc.h"
 #include "genesis/primitive-objects.h"
-#include "dynbind.h"
 
 #include <sys/types.h>
 #include <sys/time.h>
@@ -58,14 +56,13 @@
 #include "validate.h"
 #include "thread.h"
 #include "align.h"
-#include "unaligned.h"
 
 #include "gc.h"
-#include "gc-private.h"
-#include "gencgc-internal.h"
 #include <wincrypt.h>
 #include <stdarg.h>
 #include <string.h>
+#include "print.h"
+#include <shlobj.h>
 
 /* missing definitions for modern mingws */
 #ifndef EH_UNWINDING
@@ -86,126 +83,6 @@ typedef WCHAR console_char;
 typedef CHAR console_char;
 #endif
 
-/* wrappers for winapi calls that must be successful (like SBCL's
- * (aver ...) form). */
-
-/* win_aver function: basic building block for miscellaneous
- * ..AVER.. macrology (below) */
-
-/* To do: These routines used to be "customizable" with dyndebug_init()
- * and variables like dyndebug_survive_aver, dyndebug_skip_averlax based
- * on environment variables.  Those features got lost on the way, but
- * ought to be reintroduced. */
-
-static inline
-intptr_t win_aver(intptr_t value, char* comment, char* file, int line,
-                  int justwarn)
-{
-    if (!value) {
-        LPSTR errorMessage = "<FormatMessage failed>";
-        DWORD errorCode = GetLastError(), allocated=0;
-        int posixerrno = errno;
-        const char* posixstrerror = strerror(errno);
-        char* report_template =
-            "Expression unexpectedly false: %s:%d\n"
-            " ... %s\n"
-            "     ===> returned #X%p, \n"
-            "     (in thread %p)"
-            " ... Win32 thinks:\n"
-            "     ===> code %u, message => %s\n"
-            " ... CRT thinks:\n"
-            "     ===> code %u, message => %s\n";
-
-        allocated =
-            FormatMessageA(FORMAT_MESSAGE_ALLOCATE_BUFFER|
-                           FORMAT_MESSAGE_FROM_SYSTEM,
-                           NULL,
-                           errorCode,
-                           MAKELANGID(LANG_ENGLISH,SUBLANG_ENGLISH_US),
-                           (LPSTR)&errorMessage,
-                           1024u,
-                           NULL);
-
-        if (justwarn) {
-            fprintf(stderr, report_template,
-                    file, line,
-                    comment, value,
-                    get_sb_vm_thread(),
-                    (unsigned)errorCode, errorMessage,
-                    posixerrno, posixstrerror);
-        } else {
-            lose(report_template,
-                    file, line,
-                    comment, value,
-                    get_sb_vm_thread(),
-                    (unsigned)errorCode, errorMessage,
-                    posixerrno, posixstrerror);
-        }
-        if (allocated)
-            LocalFree(errorMessage);
-    }
-    return value;
-}
-
-/* sys_aver function: really tiny adaptor of win_aver for
- * "POSIX-parody" CRT results ("lowio" and similar stuff):
- * negative number means something... negative. */
-static inline
-intptr_t sys_aver(long value, char* comment, char* file, int line,
-              int justwarn)
-{
-    win_aver((intptr_t)(value>=0),comment,file,line,justwarn);
-    return value;
-}
-
-/* Check for (call) result being boolean true. (call) may be arbitrary
- * expression now; massive attack of gccisms ensures transparent type
- * conversion back and forth, so the type of AVER(expression) is the
- * type of expression. Value is the same _if_ it can be losslessly
- * converted to (void*) and back.
- *
- * Failed AVER() is normally fatal. Well, unless dyndebug_survive_aver
- * flag is set. */
-
-#define AVER(call)                                                      \
-    ({ __typeof__(call) __attribute__((unused)) me =                    \
-            (__typeof__(call))                                          \
-            win_aver((intptr_t)(call), #call, __FILE__, __LINE__, 0);      \
-        me;})
-
-/* AVERLAX(call): do the same check as AVER did, but be mild on
- * failure: print an annoying unrequested message to stderr, and
- * continue. With dyndebug_skip_averlax flag, AVERLAX stop even to
- * check and complain. */
-
-#define AVERLAX(call)                                                   \
-    ({ __typeof__(call) __attribute__((unused)) me =                    \
-            (__typeof__(call))                                          \
-            win_aver((intptr_t)(call), #call, __FILE__, __LINE__, 1);      \
-        me;})
-
-/* Now, when failed AVER... prints both errno and GetLastError(), two
- * variants of "POSIX/lowio" style checks below are almost useless
- * (they build on sys_aver like the two above do on win_aver). */
-
-#define CRT_AVER_NONNEGATIVE(call)                              \
-    ({ __typeof__(call) __attribute__((unused)) me =            \
-            (__typeof__(call))                                  \
-            sys_aver((call), #call, __FILE__, __LINE__, 0);     \
-        me;})
-
-#define CRT_AVERLAX_NONNEGATIVE(call)                           \
-    ({ __typeof__(call) __attribute__((unused)) me =            \
-            (__typeof__(call))                                  \
-            sys_aver((call), #call, __FILE__, __LINE__, 1);     \
-        me;})
-
-/* to be removed */
-#define CRT_AVER(booly)                                         \
-    ({ __typeof__(booly) __attribute__((unused)) me = (booly);  \
-        sys_aver((booly)?0:-1, #booly, __FILE__, __LINE__, 0);  \
-        me;})
-
 /* The exception handling function looks like this: */
 EXCEPTION_DISPOSITION handle_exception(EXCEPTION_RECORD *,
                                        struct lisp_exception_frame *,
@@ -234,12 +111,12 @@ static void set_seh_frame(void *frame)
 }
 #endif
 
-#if defined(LISP_FEATURE_SB_THREAD)
-
 void alloc_gc_page()
 {
-    AVER(VirtualAlloc(GC_SAFEPOINT_PAGE_ADDR, BACKEND_PAGE_BYTES,
-                      MEM_RESERVE|MEM_COMMIT, PAGE_READWRITE));
+#ifndef LISP_FEATURE_64_BIT // 64-bit uses the page below the card mark table
+    gc_assert(VirtualAlloc(GC_SAFEPOINT_PAGE_ADDR, BACKEND_PAGE_BYTES,
+                           MEM_RESERVE|MEM_COMMIT, PAGE_READWRITE));
+#endif
 }
 
 /* Permit loads from GC_SAFEPOINT_PAGE_ADDR (NB page state change is
@@ -266,19 +143,17 @@ void alloc_gc_page()
 void map_gc_page()
 {
     DWORD oldProt;
-    AVER(VirtualProtect((void*) GC_SAFEPOINT_PAGE_ADDR, BACKEND_PAGE_BYTES,
-                        PAGE_READWRITE, &oldProt));
+    gc_assert(VirtualProtect((void*) GC_SAFEPOINT_PAGE_ADDR, BACKEND_PAGE_BYTES,
+                             PAGE_READWRITE, &oldProt));
 }
 
 void unmap_gc_page()
 {
     DWORD oldProt;
-    AVER(VirtualProtect((void*) GC_SAFEPOINT_PAGE_ADDR, BACKEND_PAGE_BYTES,
-                        PAGE_NOACCESS, &oldProt));
+    gc_assert(VirtualProtect((void*) GC_SAFEPOINT_PAGE_ADDR, BACKEND_PAGE_BYTES,
+                             PAGE_NOACCESS, &oldProt));
 }
 
-#endif
-
 /* This feature has already saved me more development time than it
  * took to implement.  In its current state, ``dynamic RT<->core
  * linking'' is a protocol of initialization of C runtime and Lisp
@@ -327,7 +202,7 @@ void unmap_gc_page()
  * bundle'' that rolls up your patch, redumps and -- presto -- 100MiB
  * program is fixed by sending and loading a 50KiB thingie.
  *
- * However, until LISP_FEATURE_LINKAGE_TABLE, if your bug were fixed
+ * However, until LISP_FEATURE_ALIEN_LINKAGE_TABLE, if your bug were fixed
  * by modifying two lines of _C_ sources, a customer described above
  * had to be ready to receive and reinstall a new 100MiB
  * executable. With the aid of code below, deploying such a fix
@@ -357,15 +232,6 @@ void unmap_gc_page()
  * ``global quiesce point'' that is generally required for this kind
  * of worldwide revolution -- around collect_garbage.
  *
- * What's almost unnoticeable from the C side (where you are now, dear
- * reader): using the same style for all linking is beautiful. I tried
- * to leave old-style linking code in place for the sake of
- * _non-linkage-table_ platforms (they probably don't have -ldl or its
- * equivalent, like LL/GPA, at all) -- but i did it usually by moving
- * the entire `old style' code under #-linkage-table and
- * refactoring the `new style' branch, instead of cutting the tail
- * piecemeal and increasing #+-ifdeffery amount & the world enthropy.
- *
  * If we look at the majority of the ``new style'' code units, it's a
  * common thing to observe how #+-ifdeffery _vanishes_ instead of
  * multiplying: #-sb-xc, #+sb-xc-host and #-sb-xc-host end up
@@ -455,10 +321,10 @@ uint32_t os_get_build_time_shared_libraries(uint32_t excl_maximum,
                                        void** opt_store_handles,
                                        const char *opt_store_names[])
 {
-    void* base = opt_root ? opt_root : (void*)runtime_module_handle;
+    char* base = opt_root ? opt_root : (void*)runtime_module_handle;
     /* base defaults to 0x400000 with GCC/mingw32. If you dereference
      * that location, you'll see 'MZ' bytes */
-    void* base_magic_location =
+    char* base_magic_location =
         base + ((IMAGE_DOS_HEADER*)base)->e_lfanew;
 
     /* dos header provided the offset from `base' to
@@ -478,13 +344,13 @@ uint32_t os_get_build_time_shared_libraries(uint32_t excl_maximum,
          * fortunately, those places are irrelevant to the task at
          * hand. */
 
-        IMAGE_FILE_HEADER* image_file_header = (base_magic_location + 4);
+        IMAGE_FILE_HEADER* image_file_header = (void*)(base_magic_location + 4);
         IMAGE_OPTIONAL_HEADER* image_optional_header =
             (void*)(image_file_header + 1);
         IMAGE_DATA_DIRECTORY* image_import_direntry =
             &image_optional_header->DataDirectory[IMAGE_DIRECTORY_ENTRY_IMPORT];
         IMAGE_IMPORT_DESCRIPTOR* image_import_descriptor =
-            base + image_import_direntry->VirtualAddress;
+            (void*)(base + image_import_direntry->VirtualAddress);
         uint32_t nlibrary, j;
 
         for (nlibrary=0u; nlibrary < excl_maximum
@@ -573,7 +439,6 @@ void* os_dlsym_default(char* name)
     return result;
 }
 
-#if defined(LISP_FEATURE_SB_THREAD)
 /* We want to get a slot in TIB that (1) is available at constant
    offset, (2) is our private property, so libraries wouldn't legally
    override it, (3) contains something predefined for threads created
@@ -623,9 +488,9 @@ struct {
     CONDITION_VARIABLE cond_has_data;
     CONDITION_VARIABLE cond_has_client;
     HANDLE thread;
-    boolean initialized;
+    bool initialized;
     HANDLE handle;
-    boolean in_progress;
+    bool in_progress;
 } ttyinput;
 
 #ifdef LISP_FEATURE_64_BIT
@@ -680,7 +545,6 @@ int os_preinit(char *argv[], char *envp[])
 #endif
     return 0;
 }
-#endif  /* LISP_FEATURE_SB_THREAD */
 
 int os_number_of_processors = 1;
 
@@ -704,7 +568,7 @@ static void resolve_optional_imports()
 
 #undef RESOLVE
 
-intptr_t win32_get_module_handle_by_address(os_vm_address_t addr)
+intptr_t win32_get_module_handle_by_address(void* addr)
 {
     HMODULE result = 0;
     /* So apparently we could use VirtualQuery instead of
@@ -786,8 +650,8 @@ set_up_win64_seh_thunk(size_t page_size)
     if (page_size < sizeof(struct win64_seh_data))
         lose("Not enough space to allocate struct win64_seh_data");
 
-    AVER(VirtualAlloc(WIN64_SEH_DATA_ADDR, page_size,
-                      MEM_RESERVE | MEM_COMMIT, PAGE_EXECUTE_READWRITE));
+    gc_assert(VirtualAlloc(WIN64_SEH_DATA_ADDR, page_size,
+                           MEM_RESERVE | MEM_COMMIT, PAGE_EXECUTE_READWRITE));
 
     struct win64_seh_data *seh_data = (void *) WIN64_SEH_DATA_ADDR;
     DWORD64 base = (DWORD64) seh_data;
@@ -816,7 +680,7 @@ set_up_win64_seh_thunk(size_t page_size)
     volatile uint8_t *tramp = seh_data->handler_trampoline;
     tramp[0] = 0xFF; // jmp qword ptr [rip+2]
     tramp[1] = 0x25;
-    UNALIGNED_STORE32((tramp+2), 2);
+    UNALIGNED_STORE32((void*volatile)(tramp+2), 2);
     tramp[6] = 0x66; // 2-byte nop
     tramp[7] = 0x90;
     *(void **)(tramp+8) = handle_exception;
@@ -836,12 +700,71 @@ set_up_win64_seh_thunk(size_t page_size)
     rt->EndAddress = 16;
     rt->UnwindData = (DWORD64) ui - base;
 
-    AVER(RtlAddFunctionTable(rt, 1, base));
+    gc_assert(RtlAddFunctionTable(rt, 1, base));
 }
 #endif
 
 static LARGE_INTEGER lisp_init_time;
 static double qpcMultiplier;
+DWORD win32_page_size = 0;
+ULONG win32_stack_guarantee = 0;
+
+void
+win32_set_stack_guarantee()
+{
+    /* 64 KB appears to be enough to run the SWANK debugger + (FACT 1000). 32 KB
+     * is surprisingly not enough for the debugger. */
+    ULONG request = 64*1024;
+    if (!SetThreadStackGuarantee(&request)) {
+        fprintf(stderr, "ERROR: SetThreadStackGuarantee failed: 0x%lx.\n",
+                GetLastError());
+        fflush(stderr);
+    }
+
+    /* Store stack guarantee size the first time around so that
+     * win32_reset_stack_overflow_guard_page() may restore the stack guard. */
+    if (!win32_stack_guarantee)
+        SetThreadStackGuarantee(&win32_stack_guarantee);
+}
+
+/*
+ * The usual way to do this would be to invoke _resetstkoflw(). However, it
+ * refuses to re-establish the stack overflow guard page when the stack
+ * pointer is still within the stack area reserved by SetThreadStackGuarantee().
+ *
+ * This is invoked when SB-KERNEL:HANDLE-WIN32-EXCEPTION is unwound after
+ * handling a STACK_OVERFLOW_EXCEPTION.
+ */
+void
+win32_reset_stack_overflow_guard_page() {
+    struct thread *self = get_sb_vm_thread();
+
+    /* this is similar to CONTROL_STACK_RETURN_GUARD_PAGE on other platforms,
+     * but Windows handles the page faults on its own.
+     *
+     * From experimentation, it seems that as long as set up a guard region
+     * somewhere below the stack guarantee, Windows will manage to raise a
+     * STACK_OVERFLOW_EXCEPTION appropriately next time we exhaust the stack.
+     * Furthermore, if we reprotect sufficiently far away from the stack
+     * guarantee, user code can get away with modifying the stack while
+     * unwinding from a stack overflow condition without retriggering the guard
+     * page. See test (:EXHAUST :WRITE-TO-STACK-ON-UNWIND). */
+#define WIN32_STACK_GUARD_SLACK (2*win32_stack_guarantee + win32_page_size)
+    char *stack_guard_start = CONTROL_STACK_GUARD_PAGE(self);
+    fprintf(stderr, "INFO: Reprotecting control stack guard (0x%p+0x%lx)\n",
+            stack_guard_start, WIN32_STACK_GUARD_SLACK);
+    fflush(stderr);
+
+    DWORD oldprot;
+    VirtualProtect(stack_guard_start + WIN32_STACK_GUARD_SLACK,
+                   /* a single page would probably be enough, but let's cargo
+                    * cult Windows and use guarantee size + 1 page. */
+                   win32_stack_guarantee + win32_page_size,
+                   PAGE_READWRITE | PAGE_GUARD,
+                   &oldprot);
+
+    self->state_word.control_stack_guard_page_protected = 1;
+}
 
 void os_init()
 {
@@ -860,11 +783,9 @@ void os_init()
 
     SYSTEM_INFO system_info;
     GetSystemInfo(&system_info);
+    win32_page_size = system_info.dwPageSize;
     os_vm_page_size = system_info.dwPageSize > BACKEND_PAGE_BYTES?
         system_info.dwPageSize : BACKEND_PAGE_BYTES;
-#if defined(LISP_FEATURE_X86)
-    fast_bzero_pointer = fast_bzero_detect;
-#endif
     os_number_of_processors = system_info.dwNumberOfProcessors;
 
 #ifdef LISP_FEATURE_X86_64
@@ -886,16 +807,19 @@ uword_t get_monotonic_time()
 #endif
 
 os_vm_address_t
-os_validate(int attributes, os_vm_address_t addr, os_vm_size_t len,
-            int __attribute__((unused)) execute, int __attribute__((unused)) jit)
+os_alloc_gc_space(int space_id, int attributes, os_vm_address_t addr, os_vm_size_t len)
 {
     if (!addr) {
         int protection = attributes & IS_GUARD_PAGE ? PAGE_NOACCESS : PAGE_EXECUTE_READWRITE;
-        return
-            AVERLAX(VirtualAlloc(addr, len, MEM_RESERVE|MEM_COMMIT, protection));
+        os_vm_address_t actual = VirtualAlloc(addr, len, MEM_RESERVE|MEM_COMMIT, protection);
+        gc_assert(actual);
+        return actual;
     }
 
-    os_vm_address_t actual = VirtualAlloc(addr, len, MEM_RESERVE|MEM_COMMIT, PAGE_EXECUTE_READWRITE);
+    // Reserving the dynamic space doesn't commit it.
+    DWORD commit =
+      (space_id == DYNAMIC_CORE_SPACE_ID && (attributes & MOVABLE)) ? 0 : MEM_COMMIT;
+    os_vm_address_t actual = VirtualAlloc(addr, len, MEM_RESERVE|commit, PAGE_EXECUTE_READWRITE);
 
     if (!actual) {
         if (!(attributes & MOVABLE)) {
@@ -906,47 +830,19 @@ os_validate(int attributes, os_vm_address_t addr, os_vm_size_t len,
             return 0;
         }
 
-        return AVERLAX(VirtualAlloc(NULL, len, MEM_RESERVE|MEM_COMMIT, PAGE_EXECUTE_READWRITE));
+        actual = VirtualAlloc(NULL, len, MEM_RESERVE|commit, PAGE_EXECUTE_READWRITE);
+        gc_assert(actual);
     }
 
     return actual;
 }
 
-/* Used to allocate the dynamic space, as it may be very large. Dynamically comitted by os_commit_memory in handle_access_violation */
-os_vm_address_t
-os_validate_nocommit(int attributes, os_vm_address_t addr, os_vm_size_t len)
+void os_commit_memory(os_vm_address_t addr, os_vm_size_t len)
 {
-    os_vm_address_t actual = VirtualAlloc(addr, len, MEM_RESERVE, PAGE_EXECUTE_READWRITE);
-
-    if (!actual) {
-        if (!(attributes & MOVABLE)) {
-            fprintf(stderr,
-                    "VirtualAlloc: wanted %lu bytes at %p, actually mapped at %p\n",
-                    (unsigned long) len, addr, actual);
-            fflush(stderr);
-            return 0;
-        }
-        return AVERLAX(VirtualAlloc(NULL, len, MEM_RESERVE, PAGE_EXECUTE_READWRITE));
+    if (len) {
+        gc_assert(addr);
+        gc_assert(VirtualAlloc(addr, len, MEM_COMMIT, PAGE_EXECUTE_READWRITE));
     }
-
-    return actual;
-}
-
-void* os_commit_memory(os_vm_address_t addr, os_vm_size_t len)
-{
-    return
-        AVERLAX(VirtualAlloc(addr, len, MEM_COMMIT, PAGE_EXECUTE_READWRITE));
-}
-
-
-void os_revalidate_bzero(os_vm_address_t addr,  os_vm_size_t len) {
-    AVERLAX(VirtualFree(addr, len, MEM_DECOMMIT));
-}
-
-void
-os_invalidate(os_vm_address_t addr, os_vm_size_t len)
-{
-    AVERLAX(VirtualFree(addr, 0, MEM_RELEASE));
 }
 
 /*
@@ -958,28 +854,45 @@ os_invalidate(os_vm_address_t addr, os_vm_size_t len)
  * a lazy read (demand page) setup, but that would mean keeping an
  * open file pointer for the core indefinately (and be one more
  * thing to maintain).
+ * FIXME: I would bet that we can use PAGE_EXECUTE_WRITECOPY for this,
+ * but I'll leave it to someone who actually cares.
  */
 
 void* load_core_bytes(int fd, os_vm_offset_t offset, os_vm_address_t addr, os_vm_size_t len,
-                      int __attribute__((unused)) execute)
+                      int is_readonly_space)
 {
-    os_commit_memory(addr, len);
+    if (addr) {
+        os_commit_memory(addr, len);
+    } else {
+        addr = VirtualAlloc(NULL, len, MEM_COMMIT, PAGE_EXECUTE_READWRITE);
+        gc_assert(addr);
+    }
 #ifdef LISP_FEATURE_64_BIT
-    CRT_AVER_NONNEGATIVE(_lseeki64(fd, offset, SEEK_SET));
+    os_vm_offset_t res = _lseeki64(fd, offset, SEEK_SET);
 #else
-    CRT_AVER_NONNEGATIVE(lseek(fd, offset, SEEK_SET));
+    os_vm_offset_t res = lseek(fd, offset, SEEK_SET);
 #endif
-    size_t count;
+    gc_assert(res == offset);
+    int count;
 
+    os_vm_address_t original_addr = addr;
+    os_vm_size_t original_len = len;
     while (len) {
         unsigned to_read = len > INT_MAX ? INT_MAX : len;
         count = read(fd, addr, to_read);
+        if (count == -1) {
+            perror("read() failed"); fflush(stderr);
+        }
         addr += count;
         len -= count;
-        CRT_AVER(count == to_read);
+        gc_assert(count == (int) to_read);
     }
-    return (void*)0;
+    DWORD old;
+    if (is_readonly_space) VirtualProtect(original_addr, original_len, PAGE_READONLY, &old);
+
+    return original_addr;
 }
+
 static DWORD os_protect_modes[8] = {
     PAGE_NOACCESS,
     PAGE_READONLY,
@@ -997,16 +910,16 @@ os_protect(os_vm_address_t address, os_vm_size_t length, os_vm_prot_t prot)
     DWORD old_prot;
 
     DWORD new_prot = os_protect_modes[prot];
-    AVER(VirtualProtect(address, length, new_prot, &old_prot)||
-         (VirtualAlloc(address, length, MEM_COMMIT, new_prot) &&
-          VirtualProtect(address, length, new_prot, &old_prot)));
+    gc_assert(VirtualProtect(address, length, new_prot, &old_prot)||
+              (VirtualAlloc(address, length, MEM_COMMIT, new_prot) &&
+               VirtualProtect(address, length, new_prot, &old_prot)));
     odxprint(misc,"Protecting %p + %p vmaccess %d "
              "newprot %08x oldprot %08x",
              address,length,prot,new_prot,old_prot);
 }
 
 /* A tiny bit of interrupt.c state we want our paws on. */
-extern boolean internal_errors_enabled;
+extern int internal_errors_enabled;
 
 extern void exception_handler_wrapper();
 
@@ -1054,7 +967,6 @@ handle_breakpoint_trap(os_context_t *ctx, struct thread* self)
      * 'kind' value (eg trap_Cerror). */
     unsigned trap = *(unsigned char *)OS_CONTEXT_PC(ctx);
 
-#ifdef LISP_FEATURE_SB_THREAD
     /* Before any other trap handler: gc_safepoint ensures that
        inner alloc_sap for passing the context won't trap on
        pseudo-atomic. */
@@ -1066,7 +978,6 @@ handle_breakpoint_trap(os_context_t *ctx, struct thread* self)
         thread_interrupted(ctx);
         return 0;
     }
-#endif
 
     /* This is just for info in case the monitor wants to print an
      * approximation. */
@@ -1074,13 +985,9 @@ handle_breakpoint_trap(os_context_t *ctx, struct thread* self)
         (lispobj *)*os_context_sp_addr(ctx);
 
     WITH_GC_AT_SAFEPOINTS_ONLY() {
-#if defined(LISP_FEATURE_SB_THREAD)
         block_blockable_signals(&ctx->sigmask);
-#endif
         handle_trap(ctx, trap);
-#if defined(LISP_FEATURE_SB_THREAD)
         thread_sigmask(SIG_SETMASK,&ctx->sigmask,NULL);
-#endif
     }
 
     /* Done, we're good to go! */
@@ -1120,7 +1027,6 @@ handle_access_violation(os_context_t *ctx,
 #endif
 
     /* Safepoint pages */
-#ifdef LISP_FEATURE_SB_THREAD
     if (fault_address == (void *) GC_SAFEPOINT_TRAP_ADDR) {
         thread_in_lisp_raised(ctx);
         return 0;
@@ -1130,22 +1036,21 @@ handle_access_violation(os_context_t *ctx,
         thread_in_safety_transition(ctx);
         return 0;
     }
-#endif
 
     /* dynamic space */
     page_index_t page = find_page_index(fault_address);
-    if (page != -1
-#ifndef LISP_FEATURE_SOFT_CARD_MARKS
-        && !PAGE_WRITEPROTECTED_P(page)
-#endif
-        ) {
+#ifdef LISP_FEATURE_SOFT_CARD_MARKS
+    if (page >= 0) lose("should not get access violation in dynamic space");
+#else
+    if (page != -1 && !PAGE_WRITEPROTECTED_P(page)) {
         os_commit_memory(PTR_ALIGN_DOWN(fault_address, os_vm_page_size),
                          os_vm_page_size);
         return 0;
     }
-    if (gencgc_handle_wp_violation(fault_address)) {
+    if (gencgc_handle_wp_violation(ctx, fault_address)) {
         return 0;
     }
+#endif
 
 #ifdef LISP_FEATURE_IMMOBILE_SPACE
     extern int immobile_space_handle_wp_violation(void*);
@@ -1172,7 +1077,9 @@ signal_internal_error_or_lose(os_context_t *ctx,
      */
 
     if (internal_errors_enabled) {
-
+        /* The exception system doesn't automatically clear pending
+         * exceptions, so we lose as soon as we execute any FP
+         * instruction unless we do this first. */
         asm("fnclex");
         /* We're making the somewhat arbitrary decision that having
          * internal errors enabled means that lisp has sufficient
@@ -1180,22 +1087,18 @@ signal_internal_error_or_lose(os_context_t *ctx,
          * aren't supposed to happen during cold init or reinit
          * anyway. */
 
-#if defined(LISP_FEATURE_SB_THREAD)
         block_blockable_signals(&ctx->sigmask);
-#endif
         fake_foreign_function_call(ctx);
 
         WITH_GC_AT_SAFEPOINTS_ONLY() {
             DX_ALLOC_SAP(context_sap, ctx);
             DX_ALLOC_SAP(exception_record_sap, exception_record);
-
-#if defined(LISP_FEATURE_SB_THREAD)
             thread_sigmask(SIG_SETMASK, &ctx->sigmask, NULL);
+
+#ifdef LISP_FEATURE_X86_64
+            asm("fninit");
 #endif
 
-            /* The exception system doesn't automatically clear pending
-             * exceptions, so we lose as soon as we execute any FP
-             * instruction unless we do this first. */
             /* Call into lisp to handle things. */
             funcall2(StaticSymbolFunction(HANDLE_WIN32_EXCEPTION),
                      context_sap,
@@ -1203,9 +1106,7 @@ signal_internal_error_or_lose(os_context_t *ctx,
         }
         /* If Lisp doesn't nlx, we need to put things back. */
         undo_fake_foreign_function_call(ctx);
-#if defined(LISP_FEATURE_SB_THREAD)
         thread_sigmask(SIG_SETMASK, &ctx->sigmask, NULL);
-#endif
         /* FIXME: HANDLE-WIN32-EXCEPTION should be allowed to decline */
         return;
     }
@@ -1214,7 +1115,7 @@ signal_internal_error_or_lose(os_context_t *ctx,
             (void*)(intptr_t)exception_record->ExceptionCode);
     fprintf(stderr, "Faulting IP: %p.\n",
             (void*)(intptr_t)exception_record->ExceptionAddress);
-    if (exception_record->ExceptionCode == EXCEPTION_ACCESS_VIOLATION) {
+    if ((long int)exception_record->ExceptionCode == EXCEPTION_ACCESS_VIOLATION) {
         MEMORY_BASIC_INFORMATION mem_info;
 
         if (VirtualQuery(fault_address, &mem_info, sizeof mem_info)) {
@@ -1298,6 +1199,17 @@ handle_exception_ex(EXCEPTION_RECORD *exception_record,
     int rc;
     EXCEPTION_DISPOSITION disp = ExceptionContinueExecution;
     switch (code) {
+    case EXCEPTION_STACK_OVERFLOW:
+    {
+        void *sp = voidreg(win32_context, sp);
+        fprintf(stderr, "INFO: Caught stack overflow exception (sp=0x%p); "
+                        "proceed with caution.\n", sp);
+        fflush(stderr);
+        self->state_word.control_stack_guard_page_protected = 0;
+        rc = -1;
+        break;
+    }
+
     case EXCEPTION_ACCESS_VIOLATION:
         rc = handle_access_violation(ctx, exception_record, fault_address, self);
         if (rc && continue_search_on_unhandled_access_violation) {
@@ -1385,12 +1297,12 @@ veh(EXCEPTION_POINTERS *ep)
     }
 
     DWORD64 rip = ep->ContextRecord->Rip;
-    DWORD code = ep->ExceptionRecord->ExceptionCode;
+    long int code = ep->ExceptionRecord->ExceptionCode;
     BOOL from_lisp =
-        (rip >= READ_ONLY_SPACE_START &&
-         rip <= READ_ONLY_SPACE_END) ||
-        (rip >= DYNAMIC_SPACE_START &&
-         rip <= DYNAMIC_SPACE_START+dynamic_space_size);
+        (rip >= DYNAMIC_SPACE_START && rip < DYNAMIC_SPACE_START+dynamic_space_size) ||
+        (rip >= READ_ONLY_SPACE_START && rip < READ_ONLY_SPACE_END) ||
+        immobile_space_p(rip) ||
+        (rip >= STATIC_SPACE_START && rip < (uword_t)static_space_free_pointer);
 
     if (code == EXCEPTION_ACCESS_VIOLATION ||
         code == STATUS_HEAP_CORRUPTION ||
@@ -1438,30 +1350,6 @@ wos_install_interrupt_handlers
 #endif
 }
 
-/*
- * The stubs below are replacements for the windows versions,
- * which can -fail- when used in our memory spaces because they
- * validate the memory spaces they are passed in a way that
- * denies our exception handler a chance to run.
- */
-
-void *memmove(void *dest, const void *src, size_t n)
-{
-    if (dest < src) {
-        size_t i;
-        for (i = 0; i < n; i++) *(((char *)dest)+i) = *(((char *)src)+i);
-    } else {
-        while (n--) *(((char *)dest)+n) = *(((char *)src)+n);
-    }
-    return dest;
-}
-
-void *memcpy(void *dest, const void *src, size_t n)
-{
-    while (n--) *(((char *)dest)+n) = *(((char *)src)+n);
-    return dest;
-}
-
 char *dirname(char *path)
 {
     static char buf[PATH_MAX + 1];
@@ -1495,8 +1383,8 @@ socket_input_available(HANDLE socket, long time, long utime)
 
     FD_ZERO(&readfds);
     FD_ZERO(&errfds);
-    FD_SET(socket, &readfds);
-    FD_SET(socket, &errfds);
+    FD_SET((uword_t)socket, &readfds);
+    FD_SET((uword_t)socket, &errfds);
 
     count = select(0, &readfds, NULL, &errfds, &timeout);
     SetLastError(wsaErrno);
@@ -1516,13 +1404,11 @@ socket_input_available(HANDLE socket, long time, long utime)
         return 0;
 }
 
-#ifdef LISP_FEATURE_SB_THREAD
 /* Atomically mark current thread as (probably) doing synchronous I/O
  * on handle, if no cancellation is requested yet (and return TRUE),
  * otherwise clear thread's I/O cancellation flag and return false.
  */
-static
-boolean io_begin_interruptible(HANDLE handle)
+static bool io_begin_interruptible(HANDLE handle)
 {
     /* No point in doing it unless OS supports cancellation from other
      * threads */
@@ -1558,9 +1444,6 @@ io_end_interruptible(HANDLE handle)
       return -1;                           \
     }                                      \
     RUN_BODY_ONCE(xx, io_end_interruptible(handle))
-#else
-#define WITH_INTERRUPTIBLE_IO(handle)
-#endif
 
 int console_handle_p(HANDLE handle)
 {
@@ -1568,7 +1451,7 @@ int console_handle_p(HANDLE handle)
     return GetFileType(handle) == FILE_TYPE_CHAR &&
         GetConsoleMode(handle, &mode);
 }
-#ifdef LISP_FEATURE_SB_THREAD
+
 /*
  * (AK writes:)
  *
@@ -1663,8 +1546,7 @@ static __stdcall unsigned int tty_read_line_server(LPVOID arg)
     return 0;
 }
 
-static boolean
-tty_maybe_initialize_unlocked(HANDLE handle)
+static bool tty_maybe_initialize_unlocked(HANDLE handle)
 {
     if (!ttyinput.initialized) {
         if (!DuplicateHandle(GetCurrentProcess(),handle,
@@ -1684,10 +1566,9 @@ tty_maybe_initialize_unlocked(HANDLE handle)
     return 1;
 }
 
-boolean
-win32_tty_listen(HANDLE handle)
+bool win32_tty_listen(HANDLE handle)
 {
-    boolean result = 0;
+    bool result = 0;
     INPUT_RECORD ir;
     DWORD nevents;
     EnterCriticalSection(&ttyinput.lock);
@@ -1783,11 +1664,10 @@ unlock:
     return result;
 }
 
-boolean
-win32_maybe_interrupt_io(void* thread)
+bool win32_maybe_interrupt_io(void* thread)
 {
     struct thread *th = thread;
-    boolean done = 0;
+    bool done = 0;
 
 #ifdef LISP_FEATURE_SB_FUTEX
     if (thread_extra_data(th)->waiting_on_address)
@@ -1819,7 +1699,6 @@ win32_maybe_interrupt_io(void* thread)
     }
     return done;
 }
-#endif
 
 static const LARGE_INTEGER zero_large_offset = {.QuadPart = 0LL};
 
@@ -1931,9 +1810,10 @@ win32_unix_write(HANDLE handle, void * buf, int count)
 }
 
 
-int
-win32_unix_read(HANDLE handle, void * buf, int count)
+ssize_t
+win32_unix_read(HANDLE handle, void * buf, size_t requested_count)
 {
+    int count = requested_count > INT_MAX ? INT_MAX : requested_count;
     OVERLAPPED overlapped = {.Internal=0};
     DWORD read_bytes = 0;
     struct thread * self = get_sb_vm_thread();
@@ -2065,31 +1945,21 @@ char *os_get_runtime_executable_path()
 DWORD
 win32_wait_object_or_signal(HANDLE waitFor)
 {
-#ifdef LISP_FEATURE_SB_THREAD
     struct thread *self = get_sb_vm_thread();
     HANDLE handles[] = {waitFor, thread_private_events(self,1)};
     return
         WaitForMultipleObjects(2,handles, FALSE, INFINITE);
-#else
-    return WaitForSingleObject(waitFor, INFINITE);
-#endif
 }
 
 DWORD
 win32_wait_for_multiple_objects_or_signal(HANDLE *handles, DWORD count)
 {
-#ifdef LISP_FEATURE_SB_THREAD
     struct thread *self = get_sb_vm_thread();
     handles[count] = thread_private_events(self,1);
     return
         WaitForMultipleObjects(count + 1, handles, FALSE, INFINITE);
-#else
-    return
-        WaitForMultipleObjects(count, handles, FALSE, INFINITE);
-#endif
 }
 
-#ifdef LISP_FEATURE_SB_THREAD
 /*
  * Portability glue for win32 waitable timers.
  *
@@ -2132,7 +2002,6 @@ os_cancel_wtimer(HANDLE handle)
 {
     CancelWaitableTimer(handle);
 }
-#endif
 
 #ifdef LISP_FEATURE_SB_FUTEX
 int
@@ -2188,30 +2057,7 @@ int sb_pthr_kill(struct thread* thread, int signum)
     return 0;
 }
 
-int sigpending(sigset_t *set)
-{
-    struct extra_thread_data* data = thread_extra_data(get_sb_vm_thread());
-    *set = InterlockedCompareExchange((volatile LONG*)&data->pending_signal_set,
-                                      0, 0);
-    return 0;
-}
-
 /* Signals */
-struct sigaction signal_handlers[NSIG];
-
-/* Never called for now */
-int sigaction(int signum, const struct sigaction* act, struct sigaction* oldact)
-{
-  struct sigaction newact = *act;
-  if (oldact)
-    *oldact = signal_handlers[signum];
-  if (!(newact.sa_flags & SA_SIGINFO)) {
-      newact.sa_sigaction = (typeof(newact.sa_sigaction))newact.sa_handler;
-  }
-  signal_handlers[signum] = newact;
-  return 0;
-}
-
 int sched_yield()
 {
   /* http://stackoverflow.com/questions/1383943/switchtothread-vs-sleep1
